!+
! Module: math_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module math_utils

  use dal
  use param  
  use errorHandling

  use constants
  use ssc_misc	! for fracVal

  implicit none 

  private 

  public incompleteGammaP
  public incompleteGammaQ
  public lnGamma
  public poissonProb
  public factorial
  public chi2Prob
  public intPoissonProb
  public gaussProb
  public erf
  public erfc
  public riddersMethod
  public riddersMethod_20030324
public incompleteGammaQByAsymtotic

!logical :: testFlag = .false.

contains 

  !---------------------------------------------------------------------
  subroutine incompleteGammaP(a, x, p, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function P(a,x), defined by
  !
  !		                 x
  !		           1    /
  !		P(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               0
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the following routine.
  !
  ! This routine was cribbed from gammp() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in)  :: a, x
    real, intent(out) :: p
    integer, intent(out), optional :: status

    integer :: localStatus
    real :: q

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    p = 0.0

    if (x .lt. 0.0 .or. a .le. 0.0) then
      if (present(status)) status = -1
  return
    end if

    if (x .lt. a + 1) then ! use the series representation:
      call incompleteGammaPbySeries(a, x, p, localStatus)
    else ! use the continued-fraction representation:
      call incompleteGammaQbyFraction(a, x, q, localStatus)
      p = 1.0 - q
    end if

    if (present(status)) status = localStatus

  end subroutine incompleteGammaP

  !---------------------------------------------------------------------
  subroutine incompleteGammaQ(a, x, q, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec function gamic(a,x).
  !
  ! This routine was cribbed from gammq() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in)  :: a, x
    real, intent(out) :: q
    integer, intent(out), optional :: status

    integer :: localStatus
    real :: p

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    p = 0.0

    if (x .lt. 0.0 .or. a .le. 0.0) then
      if (present(status)) status = -1
  return
    end if

    if (x .lt. a + 1) then ! use the series representation:
      call incompleteGammaPbySeries(a, x, p, localStatus)
      q = 1.0 - p
    else ! use the continued-fraction representation:
      call incompleteGammaQbyFraction(a, x, q, localStatus)
    end if

    if (present(status)) status = localStatus

  end subroutine incompleteGammaQ

  !---------------------------------------------------------------------
  subroutine incompleteGammaPbySeries(a, x, p, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function P(a,x), defined by
  !
  !		                 x
  !		           1    /
  !		P(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               0
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the following routine.
  !
  ! The function is here evaluated from a series.
  !
  ! This routine was cribbed from gser() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in)  :: a, x
    real, intent(out) :: p
    integer, intent(out) :: status

    integer, parameter :: maxNumIterations = 100
    real :: convergeCriterion, ap, summ, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))

    p = 0.0

    if (x .lt. 0.0) then
      status = -1
  return
    else if (x .eq. 0.0) then
  return
    end if

    ap = a
    summ = 1.0 / a
    del = summ
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      ap = ap + 1.0
      del = del * x / ap
      summ = summ + del

      if (abs(del) .lt. abs(summ) * convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
!!!  return
!*** for the time being calculate a prob anyway.
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

    p = summ * exp(-x + a * log(x) - lnGammaVal)

  end subroutine incompleteGammaPbySeries

  !---------------------------------------------------------------------
  subroutine incompleteGammaQbyFraction(a, x, q, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec routine gamic(a.x).
  !
  ! The function is evaluated by a modified Lentz's continued fraction.
  !
  ! This routine was cribbed from gcf() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in)  :: a, x
    real, intent(out) :: q
    integer, intent(out) :: status

    integer, parameter :: maxNumIterations = 100
    real :: convergeCriterion, verySmallReal, b, c, d, h, an, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))
    verySmallReal = epsilon(x)

    b = x + 1.0 - a
    c = 1.0 / verySmallReal
    d = 1.0 / b
    h = d
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      an = -i * (i - a)
      b = b + 2.0
      d = an * d + b
      if (abs(d) .lt. verySmallReal) d = verySmallReal
      c = b + an / c
      if (abs(c) .lt. verySmallReal) c = verySmallReal
      d = 1.0 / d
      del = d * c
      h = h * del
      if(abs(del - 1.0) .lt. convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
  return
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

    q = h * exp(-x + a * log(x) - lnGammaVal)

  end subroutine incompleteGammaQbyFraction

  !---------------------------------------------------------------------
  subroutine incompleteGammaQByAsymtotic(a, x, q, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec function gamic(a,x).
  !
  ! This routine was cribbed from gammq() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in)  :: a, x
    real, intent(out) :: q
    integer, intent(out) :: status

    integer, parameter :: maxNumIterations = 100
    real :: convergeCriterion, ap, summ, del, oldDel, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))
print *, 'convergeCriterion:', convergeCriterion

    q = 0.0

    if (x .lt. 0.0) then
      status = -1
  return
    else if (x .eq. 0.0) then
  return
    end if

    ap = a
    summ = 1.0
    del = summ
    oldDel = del
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      ap = ap - 1.0
      del = del * ap / x
      summ = summ + del
print *, 'del, summ:', del, summ

      if (abs(del) < abs(summ) * convergeCriterion) then
        notConverged = .false.
    exit
      end if

    if (abs(del) > abs(oldDel)) exit

      oldDel = del
    end do

    if (notConverged) then
      status = -2
!!!  return
!*** for the time being calculate a prob anyway.
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

    q = summ * exp(-x + (a - 1) * log(x) - lnGammaVal)
print *, 'q:', q

  end subroutine incompleteGammaQByAsymtotic

  !---------------------------------------------------------------------
  real function lnGamma(a)
  !---------------------------------------------------------------------
  ! Returns the natural log of the absolute value of the complete gamma
  ! function T(a), which is defined by
  !
  !		       inf
  !		       /
  !		T(a) = | dt exp(-t) t^(a-1).
  !		       /
  !		      0
  !
  ! The function is evaluated by a Lanczos approximation.
  !
  ! The ln|T(a)| function can also be evaluated using the slatec function
  ! alngam(a).
  !
  ! This routine was cribbed from gammln() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in) :: a

    integer :: j
    real(kind(1d0)) :: ser, stp, tmp, x, y, cof(6)

    save cof, stp
    data cof, stp /&
      76.18009172947146d0,&
     -86.50532032941677d0,&
      24.01409824083091d0,&
      -1.231739572450155d0,&
       0.1208650973866179d-2,&
      -0.5395239384953d-5,&
       2.5066282746310005d0/

    !-------------------------------------------------------------------
    ! Start

    x = a
    y = x
    tmp = x + 5.5d0
    tmp = (x + 0.5d0) * log(tmp) - tmp
    ser = 1.000000000190015d0
    do j = 1, 6
      y = y + 1.0d0
      ser = ser + cof(j) / y
    end do
    lnGamma = tmp + log(stp * ser / x)

  end function lnGamma

  !---------------------------------------------------------------------
  real function poissonProb(av, i)
  !---------------------------------------------------------------------
  ! Returns the Poisson probability of the occurrence of i events when
  ! the ensemble average = av. This probability distribution is given by
  !
  !	        av^i exp(-av)
  !	P(i) = ---------------
  !	             i!
  !
  ! This routine was adapted from material in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real, intent(in) :: av
    integer, intent(in) :: i

    logical :: firstEntry = .true.
    integer, save :: threshhold
    real :: logHugeInt, nOld, nNew

    !-------------------------------------------------------------------
    ! Start

    poissonProb = 0.0

  if (av < 0.0) return
  if (i  < 0) return

    if (av == 0.0 .and. i == 0) then
      poissonProb = 1.0
  return
    end if

    if (firstEntry) then ! calculate the threshhold value corresponding to
    ! the max integer size of the present platform:

      firstEntry = .false.

      ! Invert Stirling's formula for factorial(n) by iteration:
      logHugeInt = log(real(huge(i)))
      nOld = logHugeInt
      do
        nNew = (logHugeInt - 0.5 * log(TwoPi * nOld)) / (log(nOld) - 1.0)
      if (abs(nNew - nOld) < 1.0) exit
        nOld = nNew
      end do

      threshhold = floor(nNew - 1) ! the minus 1 just for a bit of elbow room.
    end if

    if (i < threshhold) then
      poissonProb = exp(i * log(av) - av) / factorial(i)

    else
      poissonProb = exp(i * log(av) - av - lnGamma(i + 1.0))
    end if

  end function poissonProb

  !---------------------------------------------------------------------
  integer function factorial(i)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: i

    integer :: j

    !-------------------------------------------------------------------
    ! Start

    factorial = 1
    do j = 2, i
      factorial = factorial * j
    end do

  end function factorial

  !---------------------------------------------------------------------
  subroutine chi2Prob(chi2, degFree, probability, status)
  !---------------------------------------------------------------------
  ! This subroutine gives the probability that a given chi2/degFree is
  ! 'natural' - ie it gives the probability of the null hypothesis. Chi2
  ! (chi squared) is defined as follows:
  !
  !	         sum[i=1, N](y_i - e_i)^2
  !	chi^2 = --------------------------
  !	                 sigma_e_i^2
  !
  ! where y_i is the measured or data value, e_i is the model or expected
  ! value and sigma_e_i is the uncertainty in the latter.

    implicit none

    real,    intent(in)            :: chi2
    integer, intent(in)            :: degFree
    real,    intent(out)           :: probability
    integer, intent(out), optional :: status

    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    probability = 0.0

    if (chi2 .le. 0.0) then
      if (present(status)) status = -1
  return
    end if
    if (degFree .le. 0) then
      if (present(status)) status = -2
  return
    end if

    call incompleteGammaQ(degFree / 2.0, chi2 / 2.0, probability, foreignStatus)

    if (present(status)) status = foreignStatus

  end subroutine chi2Prob

  !---------------------------------------------------------------------
  real function intPoissonProb(av, i)
  !---------------------------------------------------------------------
  ! Returns the probability that a Poisson-distributed (integer) variable
  ! i will EQUAL OR exceed av. (Note that this is slightly different
  ! from the corresponding function in simutils/math_utils.

    implicit none

    real, intent(in) :: av
    integer, intent(in) :: i

    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    if (av < 0.0) then
      intPoissonProb = 0.0
  return
    end if

    if (i <= 0) then
      intPoissonProb = 1.0
  return
    end if

    call incompleteGammaP(real(i), av, intPoissonProb, foreignStatus)

    if (foreignStatus .ne. 0) intPoissonProb = 0.0

  end function intPoissonProb

  !---------------------------------------------------------------------
  subroutine intPoissonProb_old(testY, meanY, probability, status)
  !---------------------------------------------------------------------
  ! Returns the probability that a Poisson-distributed (integer) variable
  ! y will exceed testY.

    implicit none

    integer, intent(in) :: testY
    real, intent(in) :: meanY
    real, intent(out) :: probability
    integer, intent(out) :: status

    !-------------------------------------------------------------------
    ! Start

    status = 0
    probability = 0.0
    if (testY .lt. 0) then
      status = -1
  return
    end if
    if (meanY .lt. 0.0) then
      status = -2
  return
    end if

    call incompleteGammaP(real(testY + 1), meanY, probability, status)

  end subroutine intPoissonProb_old

  !---------------------------------------------------------------------
  function gaussProb(testY, meanY, sigma) result(probability)
  !---------------------------------------------------------------------
  ! Returns the probability that a gaussian-distributed variable y will
  ! depart from the mean by greater than abs(testY - meanY). For values
  ! obeying a gaussian distribution of given mean and
  ! standard deviation s, the probability of a given y value (or greater)
  ! occurring by chance is 1 - erf(abs(testY - meanY)/s/sqrt(2))

    implicit none

    real, intent(in) :: testY, meanY, sigma
    real :: probability

    !-------------------------------------------------------------------
    ! Start

!    probability = 1.0 - erf(real(abs(testY - meanY) / sigma / RootTwo))
    probability = erfc(real(abs(testY - meanY) / sigma / RootTwo))

  end function gaussProb

  !---------------------------------------------------------------------
  real function erf(x, style)
  !---------------------------------------------------------------------
  ! The error function
  !
  !	                     x
  !	             2      /
  !	erf(x) = ---------- | dt exp(-t^2).
  !	          sqrt(pi)  /
  !	                   0
  !

    implicit none

    real, intent(in) :: x
    integer, intent(in), optional :: style

    integer :: localStyle

    !-------------------------------------------------------------------
    ! Start

    if (present(style)) then
      localStyle = style
    else
      localStyle = 0
    end if

    select case(localStyle)
    case(0)
      call incompleteGammaP(0.5, x*x, erf)
    case(1)
      erf = erfApprox0(x)
    case(2)
      erf = erfApprox1(x)
    case default
    end select

  end function erf

  !---------------------------------------------------------------------
  real function erfc(x, style)
  !---------------------------------------------------------------------
  ! The complementary error function
  !
  !	                                   inf
  !	                           2      /
  !	erfc(x) = 1 - erf(x) = ---------- | dt exp(-t^2).
  !	                        sqrt(pi)  /
  !	                                 x
  !
  ! Note that, for values obeying a gaussian distribution of mean <y> and
  ! standard deviation s, the probability of a given y value (or greater)
  ! occurring by chance is erfc(abs(y - <y>)/s/sqrt(2))

    implicit none

    real, intent(in) :: x
    integer, intent(in), optional :: style

    integer :: localStyle

    !-------------------------------------------------------------------
    ! Start

    if (present(style)) then
      localStyle = style
    else
      localStyle = 0
    end if

    select case(localStyle)
    case(0)
      call incompleteGammaQ(0.5, x*x, erfc)
    case(1)
      erfc = 1.0 - erfApprox0(x)
    case(2)
      erfc = 1.0 - erfApprox1(x)
    case default
    end select

  end function erfc

  !---------------------------------------------------------------------
  function erfApprox0(x) result(erf)
  !---------------------------------------------------------------------
  ! Taken from Numerical Methods for Fortran.

    implicit none

    real, intent(in) :: x

    real :: t, z, erf
    real, parameter ::&
      p0 = -1.26551223,&
      p1 =  1.00002368,&
      p2 =  0.37409196,&
      p3 =  0.09678418,&
      p4 = -0.18628806,&
      p5 =  0.27886807,&
      p6 = -1.13520398,&
      p7 =  1.48851587,&
      p8 = -0.82215223,&
      p9 =  0.17087277

    !-------------------------------------------------------------------
    ! Start

    z = abs(x)
    t = 1.0 / (1.0 + 0.5 * z)
    erf = t * exp(-z * z + p0 + t * (p1 + t * (p2&
        + t * (p3 + t * (p4 + t * (p5 + t * (p6&
        + t * (p7 + t * (p8 + t * p9)))))))))

    if (x .lt. 0.0) erf = 2.0 - erf

  end function erfApprox0

  !---------------------------------------------------------------------
  function erfApprox1(x) result(erf)
  !---------------------------------------------------------------------
  ! This approx uses the formula of Busen K M, Am J Phys 49, 886 (1981).

    implicit none

    real, intent(in) :: x
    real :: erf, sign

    !-------------------------------------------------------------------
    ! Start

    sign = 1.0
    if (x < 0.0) sign = -1.0

    if (x > 1.0) then
      erf = sqrt(1.0 - exp(-4.0 * x**1.88 / Pi))
    else
      erf = sqrt(1.0 - exp(-4.0 * x * x / Pi))
    end if

    erf = erf * sign

  end function erfApprox1

  !---------------------------------------------------------------------
  function erfcApprox0(x) result(erfc)
  !---------------------------------------------------------------------
  !x    input	Deviate, range -infinity to +infinity.
  !erfc output	Area under Gaussian from X to +infinity.
  !-
  ! Author	Clive Page	1983 July 5
  ! Standard Fortran-77 version 1984 Sept 25, CGP.
  ! From algorithm AS66 by I.D.Hill, Appl Stats  22,424 (1973).
  ! Adapted to f90 by Ian Stewart 8 Aug 2000

    implicit none 

    real, intent(in) :: x
    real :: erfc, y, z

    !-------------------------------------------------------------------
    ! Start

    z = abs(x)
    if(z .gt. 13.0) then
      erfc = 0.0

    else
      y = 0.5 * z * z
      if(z .le. 1.28) then
        erfc = 0.5 - z * (0.398942280444 - 0.399903438504 * y /&
        (y + 5.75885480458 - 29.8213557808 /&
        (y + 2.62433121679 + 48.6959930692 /&
        (y + 5.92885724438))))

      else
        erfc = 0.398942280385 * exp(-y) /&
        (z - 3.8052E-8      + 1.00000615302 /&
        (z + 3.98064794E-4  + 1.98615381364 /&
        (z - 0.151689116635 + 5.29330324926 /&
        (z + 4.8385912808   - 15.1508972451 /&
        (z + 0.742380924027 + 30.789933034  /&
        (z + 3.99019417011))))))
      end if
    end if

    if(x .lt. 0.0) erfc = 1.0 - erfc

  end function erfcApprox0

  !---------------------------------------------------------------------
  subroutine riddersMethod(startX0, startX1, y, solutionX&
  , maxNumIterations, convergenceCriterion, dx, status)
  !---------------------------------------------------------------------
  ! This subroutine has been adapted from an algorithm described in
  ! Numerical Methods for Fortran. I changed it slightly in that the
  ! convergence test compares the difference between the present and
  ! last values of xNew, rather than the difference between xHi and xLo
  ! as in the original algorithm, to the value of convergenceCriterion.
  !
  ! This subroutine solves the equation y(x) = 0 by Ridders' Method. The
  ! user must construct the function y(x) in the calling module. This
  ! function (and its single argument x) must be of type real. The user
  ! should pass the function name in the call to riddersMethod.
  !
  ! Note that y(startX0) and y(startX1) must be of opposite sign.

    implicit none

    real,    intent(in)            :: startX0, startX1
    real,    intent(out)           :: solutionX
    integer, intent(in),  optional :: maxNumIterations
    real,    intent(in),  optional :: convergenceCriterion
    real,    intent(out), optional :: dx
    integer, intent(out), optional :: status

    interface
      real function y(x)
        real, intent(in) :: x
      end function y
    end interface

    integer ::&
      localStatus, j
    real ::&
      startY0, startY1, xLo, yLo, xMid, yMid, xHi, yHi, xOld, xNew, yNew,&
      root, expQ, fracY, localConvergenceCriterion

    !-------------------------------------------------------------------
    ! Start

!print *, ''
    localStatus = 0 ! default
    if (present(status)) status = localStatus
    solutionX = 0.0 ! Don't like to leave it with no value, even if the routine
    ! can't calculate it.
    dx = 0.0 ! default

!print *, ' >>  '
!print *, ' >>  Calc startY0:'
    startY0 = y(startX0)
!print *, ' >>  Calc startY1:'
    startY1 = y(startX1)

!print *, ' >> startY0, startY1:', startY0, startY1

    if (startY0 == 0) then
      solutionX = startX0
  return
    end if
    if (startY1 == 0) then
      solutionX = startX1
  return
    end if

    if (startX0 == startX1) then
      if (present(status)) status = -1 ! no space between startX0 and startX1.
  return
    end if

    if (startY0 > 0.0 .and. startY1 > 0.0&
    .or. startY0 < 0.0 .and. startY1 < 0.0) then
!print *, '>> startX0, startX1, startY0, startY1: ', startX0, startX1, startY0, startY1
      if (present(status)) status = -2
  return
    end if

    ! Orient bounds such that y(xLo) < 0:
    if (startY0 < 0.0) then
      xLo = startX0
      yLo = startY0
      xHi = startX1
      yHi = startY1
    else
      xLo = startX1
      yLo = startY1
      xHi = startX0
      yHi = startY0
    end if

    dx = xHi - xLo
xOld = xLo ! could have also chosen xHi; at this point it is indeterminate which of these two values is the 'latest estimate'.

    if (present(convergenceCriterion)) then
      localConvergenceCriterion = abs(convergenceCriterion)
!print *, 'xLo, xHi, localConvergenceCriterion:', xLo, xHi, localConvergenceCriterion
    else
      localConvergenceCriterion = 0.5 * abs(startX0 + startX1)&
      * (10.0 ** (1 - precision(dx)))
!print *, 'WRONG!'
    end if

!print *, ' >>  xLo, xHi: ', xLo, xHi
!print *, ' >>  yLo, yHi: ', yLo, yHi

!print *, ' >>  Commence loop:'
    j = 1
    do
!print *, '  >>  j, maxNumIterations: ', j, maxNumIterations
      xMid = 0.5 * (xLo + xHi)
!print *, '  >>  Calc yMid:'
      yMid = y(xMid)
!print *, '  >>  yMid=', yMid
!print *, '  >>  xLo, xMid, xHi: ', xLo, xMid, xHi
!print *, '  >>  yLo, yMid, yHi: ', yLo, yMid, yHi
      root = sqrt(yMid**2 - yLo * yHi)
!print *, '  >>  root = sqrt(yMid**2 - yLo * yHi)=', root

      if (root == 0) then ! The argument of the square root can
      ! only equal zero if one at least of yLo and yHi equals zero (and
      ! if yMid is also equal to zero). The routine is supposed to screen out
      ! such cases before this point is reached. This status value indicates
      ! a bug.
        if (present(status)) status = -3
  return
      end if

      ! Solve quadratic equation to get expQ:

      expQ = (yMid + root) / yHi ! That yHi > 0 is checked implicitly
      ! elsewhere in the loop.
!print *, '  >>  expQ = (yMid + root) / yHi =', expQ

      ! Apply false-position method to y(x)*exp(2Q(x-xLo)/(xHi-xLo)):

      fracY = -yLo / (yHi * expQ * expQ - yLo)
      xNew = fracVal(xLo, xHi, fracY)
!print *, '  >>  Calc yNew:'
      yNew = y(xNew)
!print *, '  >>  fracY = -yLo / (yHi * expQ * expQ - yLo) =', fracY
!print *, '  >>  xNew = fracVal(xLo, xHi, fracY) =', xNew
!print *, '  >>  yNew = ', yNew

      if (yNew == 0.0) then
        solutionX = xNew
!print *, '  >>  exit because yNew == 0.0'
    exit
      end if

      ! Now decide new x bounds:

      if (yNew < 0.0) then
        xLo = xNew
        yLo = yNew
      else
        xHi = xNew
        yHi = yNew
      end if
!print *, '  >>  new xLo, xHi: ', xLo, xHi
!print *, '  >>  new yLo, yHi: ', yLo, yHi

      ! Check convergence:

!      dx = xHi - xLo
dx = xNew - xOld
!print *, '  >>  xNew, xOld, abs(dx=xNew-xOld), localConvergenceCriterion: '&
!, xNew, xOld, abs(dx), localConvergenceCriterion
      if (abs(dx) < localConvergenceCriterion) then
        solutionX = xNew
!print *, '  >>  exit because abs(dx) < localConvergenceCriterion'
    exit
      end if

      j = j + 1
      if (present(maxNumIterations)) then
        if (j > maxNumIterations) then
          localStatus = ibset(localStatus, 0)
          solutionX = xNew
!print *, '  >>  exit because j+1 > maxNumIterations'
    exit
        end if
      end if

xOld = xNew
    end do

    if (present(status)) status = localStatus
!print *, ''

  end subroutine riddersMethod

  !---------------------------------------------------------------------
  subroutine riddersMethod_20030324(startX0, startX1, y, solutionX&
  , maxNumIterations, convergenceCriterion, dx, status)
  !---------------------------------------------------------------------
  ! This subroutine has been adapted from an algorithm described in
  ! Numerical Methods for Fortran.
  !
  ! This subroutine solves the equation y(x) = 0 by Ridders' Method. The
  ! user must construct the function y(x) in the calling module. This
  ! function (and its single argument x) must be of type real. The user
  ! should pass the function name in the call to riddersMethod.
  !
  ! Note that y(startX0) and y(startX1) must be of opposite sign.

    implicit none

    real,    intent(in)            :: startX0, startX1
    real,    intent(out)           :: solutionX
    integer, intent(in),  optional :: maxNumIterations
    real,    intent(in),  optional :: convergenceCriterion
    real,    intent(out), optional :: dx
    integer, intent(out), optional :: status

    interface
      real function y(x)
        real, intent(in) :: x
      end function y
    end interface

    integer ::&
      localStatus, j
    real ::&
      startY0, startY1, xLo, yLo, xMid, yMid, xHi, yHi, xNew, yNew,&
      root, expQ, fracY, localConvergenceCriterion

    !-------------------------------------------------------------------
    ! Start

!print *, ''
    localStatus = 0 ! default
    if (present(status)) status = localStatus
    solutionX = 0.0 ! Don't like to leave it with no value, even if the routine
    ! can't calculate it.
    dx = 0.0 ! default

!print *, ' >>  '
!print *, ' >>  Calc startY0:'
    startY0 = y(startX0)
!print *, ' >>  Calc startY1:'
    startY1 = y(startX1)

!print *, ' >> startY0, startY1:', startY0, startY1

    if (startY0 == 0) then
      solutionX = startX0
  return
    end if
    if (startY1 == 0) then
      solutionX = startX1
  return
    end if

    if (startX0 == startX1) then
      if (present(status)) status = -1 ! no space between startX0 and startX1.
  return
    end if

    if (startY0 > 0.0 .and. startY1 > 0.0&
    .or. startY0 < 0.0 .and. startY1 < 0.0) then
!print *, '>> startX0, startX1, startY0, startY1: ', startX0, startX1, startY0, startY1
      if (present(status)) status = -2
  return
    end if

    ! Orient bounds such that y(xLo) < 0:
    if (startY0 < 0.0) then
      xLo = startX0
      yLo = startY0
      xHi = startX1
      yHi = startY1
    else
      xLo = startX1
      yLo = startY1
      xHi = startX0
      yHi = startY0
    end if

    dx = xHi - xLo

    if (present(convergenceCriterion)) then
      localConvergenceCriterion = abs(convergenceCriterion)
!print *, 'xLo, xHi, localConvergenceCriterion:', xLo, xHi, localConvergenceCriterion
    else
      localConvergenceCriterion = 0.5 * abs(startX0 + startX1)&
      * (10.0 ** (1 - precision(dx)))
!print *, 'WRONG!'
    end if

!print *, ' >>  xLo, xHi: ', xLo, xHi
!print *, ' >>  yLo, yHi: ', yLo, yHi

!print *, ' >>  Commence loop:'
    j = 1
    do
!print *, '  >>  j, maxNumIterations: ', j, maxNumIterations
      xMid = 0.5 * (xLo + xHi)
!print *, '  >>  Calc yMid:'
      yMid = y(xMid)
!print *, '  >>  yMid=', yMid
!print *, '  >>  xLo, xMid, xHi: ', xLo, xMid, xHi
!print *, '  >>  yLo, yMid, yHi: ', yLo, yMid, yHi
      root = sqrt(yMid**2 - yLo * yHi)
!print *, '  >>  root = sqrt(yMid**2 - yLo * yHi)=', root

      if (root == 0) then ! The argument of the square root can
      ! only equal zero if one at least of yLo and yHi equals zero (and
      ! if yMid is also equal to zero). The routine is supposed to screen out
      ! such cases before this point is reached. This status value indicates
      ! a bug.
        if (present(status)) status = -3
  return
      end if

      ! Solve quadratic equation to get expQ:

      expQ = (yMid + root) / yHi ! That yHi > 0 is checked implicitly
      ! elsewhere in the loop.
!print *, '  >>  expQ = (yMid + root) / yHi =', expQ

      ! Apply false-position method to y(x)*exp(2Q(x-xLo)/(xHi-xLo)):

      fracY = -yLo / (yHi * expQ * expQ - yLo)
      xNew = fracVal(xLo, xHi, fracY)
!print *, '  >>  Calc yNew:'
      yNew = y(xNew)
!print *, '  >>  fracY = -yLo / (yHi * expQ * expQ - yLo) =', fracY
!print *, '  >>  xNew = fracVal(xLo, xHi, fracY) =', xNew
!print *, '  >>  yNew = ', yNew

      if (yNew == 0.0) then
        solutionX = xNew
!print *, '  >>  exit because yNew == 0.0'
    exit
      end if

      ! Now decide new x bounds:

      if (yNew < 0.0) then
        xLo = xNew
        yLo = yNew
      else
        xHi = xNew
        yHi = yNew
      end if
!print *, '  >>  new xLo, xHi: ', xLo, xHi
!print *, '  >>  new yLo, yHi: ', yLo, yHi

      ! Check convergence:

      dx = xHi - xLo
!print *, '  >>  abs(dx=xHi-xLo), localConvergenceCriterion: ', abs(dx), localConvergenceCriterion
      if (abs(dx) < localConvergenceCriterion) then
        solutionX = xNew
!print *, '  >>  exit because abs(dx) < localConvergenceCriterion'
    exit
      end if

      j = j + 1
      if (present(maxNumIterations)) then
        if (j > maxNumIterations) then
          localStatus = ibset(localStatus, 0)
          solutionX = xNew
!print *, '  >>  exit because j+1 > maxNumIterations'
!localStatus=-10
    exit
        end if
      end if
    end do

    if (present(status)) status = localStatus
!print *, ''

  end subroutine riddersMethod_20030324

!  !---------------------------------------------------------------------
!  subroutine modifiedSecant(startX, y, solutionX&
!  , maxNumIterations, convergenceCriterion, status)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    real,    intent(in)            :: startX
!    real,    intent(out)           :: solutionX
!    integer, intent(in),  optional :: maxNumIterations
!    real,    intent(in),  optional :: convergenceCriterion
!    integer, intent(out), optional :: status
!
!    interface
!      real function y(x)
!        real, intent(in) :: x
!      end function y
!    end interface
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    localStatus = 0 ! default
!
!    if (present(status)) status = localStatus
!
!  end subroutine modifiedSecant

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module math_utils
