!
! Module: save_image
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines for dealing with keywords and columns of FITS files.
! 
! Comments: TODO:
!*********** Look through ALL software for unnecessary pointer allocations!
!*********** Look through ALL software and declare all ptrs => null()!
! 
! $Id: save_image.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module save_image

  use dal

use wcs_aux
use fits_utils

  implicit none 

  private 

  public saveOutImage

  interface saveOutImage
    module procedure saveOutImageReal32, saveOutImageInt32, saveOutImageBoolean
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine saveOutImageReal32(newImage, oldImageSetName, newImageSetName&
  , xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, loads a new
  ! REAL32 array, copies all keywords from the 0th array, deletes the
  ! 0th array, then adjusts wcs keywords if present to conform to the
  ! start pixels xStartArg and yStartArg of the new image.

    implicit none

    real(single),   intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    integer(int32), pointer ::&
      newDimVector(:)
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    real(single), pointer ::&
      arrayDataReal32(:,:)

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    allocate(newDimVector(0:1))
    newDimVector(0) = size(newImage, 1)
    newDimVector(1) = size(newImage, 2)

    call saveOutImageCommon(oldImageSetName, newImageSetName&
      , xStart, yStart, newDimVector, REAL32, set, newImageArray)

    arrayDataReal32 => real32Array2Data(newImageArray)
    arrayDataReal32 = newImage

    call release(set)
    deallocate(newDimVector)

  end subroutine saveOutImageReal32

  !---------------------------------------------------------------------
  subroutine saveOutImageBoolean(newImage, oldImageSetName, newImageSetName&
  , xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, loads a new
  ! REAL32 array, copies all keywords from the 0th array, deletes the
  ! 0th array, then adjusts wcs keywords if present to conform to the
  ! start pixels xStartArg and yStartArg of the new image.
  !
  ! The output array is actually int8 rather than boolean.

    implicit none

!    logical(bool), intent(in)           :: newImage(:,:)
logical, intent(in)           :: newImage(:,:)
    character(*),     intent(in)           :: oldImageSetName,&
                                              newImageSetName
    integer(int16),   intent(in), optional :: xStartArg,&
                                              yStartArg

    integer(int16) ::&
      xStart, yStart
    integer(int32), pointer ::&
      newDimVector(:)
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:)

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    allocate(newDimVector(0:1))
    newDimVector(0) = size(newImage, 1)
    newDimVector(1) = size(newImage, 2)

    call saveOutImageCommon(oldImageSetName, newImageSetName&
      , xStart, yStart, newDimVector, INTEGER8, set, newImageArray)

    arrayDataInt8 => int8Array2Data(newImageArray)
    arrayDataInt8 = 0
    where(newImage) arrayDataInt8 = 1

    call release(set)
    deallocate(newDimVector)

  end subroutine saveOutImageBoolean

  !---------------------------------------------------------------------
  subroutine saveOutImageCommon(oldImageSetName, newImageSetName&
  , xStart, yStart, newDimVector, newDataType, newSet, newArray)
  !---------------------------------------------------------------------
!*** The previous version seemed to be writing the new array to the primary HDU and tasks that use the dal could not open it. Let's try another way:

    implicit none

    character(*),   intent(in)  :: oldImageSetName,&
                                   newImageSetName
    integer(int16), intent(in)  :: xStart,&
                                   yStart
    integer(int32), intent(in)  :: newDimVector(:), newDataType
    type(DataSetT), intent(out) :: newSet ! just so it can be released
    type(ArrayT),   intent(out) :: newArray

    type(DataSetT) :: refSet
    type(ArrayT) :: refArray
    integer :: foreignStatus
    type(wcsDataType) :: setWcs, wcs
    logical :: wcsPresent = .false. ! default

    !-------------------------------------------------------------------
    ! Start

    if (oldImageSetName .eq. newImageSetName) then
      newSet = dataSet(oldImageSetName, MODIFY)
      refSet = clone(oldImageSetName, 'tempset.ds', TEMP)
!**** NOTE should pass this temp set name as an argument, to allow it to be specified to the task on the command line.

    else
      refSet = dataSet(oldImageSetName, READ)
      newSet = clone(oldImageSetName, newImageSetName, MODIFY)
    end if

    call deleteBlock(newSet, 0)
!*** allow for input of name or block number??

call release(newSet)
newSet = dataSet(newImageSetName, MODIFY)

    newArray=addArray(newSet, "PRIMARY", newDataType, newDimVector)
    refArray = array(refSet, 0, READ)
    call setAttributes(attributable(newArray), attributable(refArray))
    call release(refSet)

    ! Regularize and adjust the wcs:

    call getWcsInfo(newSet, setWcs, foreignStatus)
    if (foreignStatus == 0) then ! set contains the wcs keywords. Check
    ! to see if the array block also has a set:

      call getWcsInfo(newArray, wcs, foreignStatus)
      if (foreignStatus == 0) then ! array also contains the wcs keywords.
        call warning('setAlsoHasWcs','')

      else ! array doesn't contains any wcs keywords. Copy over those from
      ! the set:
        call setAttribute(newArray, 'CRPIX1', setWcs%X%refPixels)
        call setAttribute(newArray, 'CRPIX2', setWcs%Y%refPixels)
        call setAttribute(newArray, 'CDELT1', setWcs%X%pixelSizeWorld)
        call setAttribute(newArray, 'CDELT2', setWcs%Y%pixelSizeWorld)
        call setAttribute(newArray, 'CRVAL1', setWcs%X%refWorld)
        call setAttribute(newArray, 'CRVAL2', setWcs%Y%refWorld)
        call setAttribute(newArray, 'CTYPE1', setWcs%X%ctype)
        call setAttribute(newArray, 'CTYPE2', setWcs%Y%ctype)

        call warning('setAlsoHasWcs','')
      end if

      wcsPresent = .true.

    else ! set header doesn't contain any wcs keywords. Check the array:
      call getWcsInfo(newArray, wcs, foreignStatus)
      if (foreignStatus .ne. 0) wcsPresent = .true.
    end if

    if (wcsPresent) then
      if (xStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX1', wcs%X%refPixels - (xStart - 1))
      end if
      if (yStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX2', wcs%Y%refPixels - (yStart - 1))
      end if
!*** all this assumes that the pixel sizes stay the same!
    end if

  end subroutine saveOutImageCommon

  !---------------------------------------------------------------------
  subroutine saveOutImageCommon_20030415(oldImageSetName, newImageSetName&
  , xStart, yStart, newDimVector, newDataType, set, newArray)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in)  :: oldImageSetName,&
                                   newImageSetName
    integer(int16), intent(in)  :: xStart,&
                                   yStart
    integer(int32), intent(in)  :: newDimVector(:), newDataType
    type(DataSetT), intent(out) :: set ! just so it can be released
    type(ArrayT),   intent(out) :: newArray

    type(ArrayT) :: oldArray
    character(255) :: oldArrayName, tempOldArrayName
    integer :: foreignStatus
    type(wcsDataType) :: setWcs, wcs
    logical :: wcsPresent = .false. ! default

    !-------------------------------------------------------------------
    ! Start

    if (oldImageSetName .eq. newImageSetName) then
      set = dataSet(oldImageSetName, MODIFY)
    else
      set = clone(oldImageSetName, newImageSetName, MODIFY)
    end if

    oldArray = array(set, 0, READ)
!*** allow for input of name or block number??

    oldArrayName = name(oldArray)
    tempOldArrayName = trim(oldArrayName)//'_TEMP'
if (hasBlock(set, tempOldArrayName)) call error('','')
    call rename(oldArray, tempOldArrayName)
    newArray = addArray(set, trim(oldArrayName), newDataType, newDimVector)
!***** NOTE! I have had problems with this in low memory model.
    call setAttributes(attributable(newArray), attributable(oldArray))
    call deleteBlock(set, tempOldArrayName)

    call getWcsInfo(set, setWcs, foreignStatus)
    if (foreignStatus == 0) then ! set contains the wcs keywords. Check
    ! to see if the array block also has a set:

      call getWcsInfo(newArray, wcs, foreignStatus)
      if (foreignStatus == 0) then ! array also contains the wcs keywords.
        call warning('setAlsoHasWcs','')

      else ! array doesn't contains any wcs keywords. Copy over those from
      ! the set:
        call setAttribute(newArray, 'CRPIX1', setWcs%X%refPixels)
        call setAttribute(newArray, 'CRPIX2', setWcs%Y%refPixels)
        call setAttribute(newArray, 'CDELT1', setWcs%X%pixelSizeWorld)
        call setAttribute(newArray, 'CDELT2', setWcs%Y%pixelSizeWorld)
        call setAttribute(newArray, 'CRVAL1', setWcs%X%refWorld)
        call setAttribute(newArray, 'CRVAL2', setWcs%Y%refWorld)
        call setAttribute(newArray, 'CTYPE1', setWcs%X%ctype)
        call setAttribute(newArray, 'CTYPE2', setWcs%Y%ctype)

        call warning('setAlsoHasWcs','')
      end if

      wcsPresent = .true.

    else ! set header doesn't contain any wcs keywords. Check the array:
      call getWcsInfo(newArray, wcs, foreignStatus)
      if (foreignStatus .ne. 0) wcsPresent = .true.
    end if

    if (wcsPresent) then
      if (xStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX1', wcs%X%refPixels - (xStart - 1))
      end if
      if (yStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX2', wcs%Y%refPixels - (yStart - 1))
      end if
!*** all this assumes that the pixel sizes stay the same!
    end if

  end subroutine saveOutImageCommon_20030415

  !---------------------------------------------------------------------
  subroutine saveOutImageReal32_old(newImage, imageName, newImageName&
  , xStartArg, yStartArg)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) ::&
      newImage(:,:)
    character(*), intent(in) ::&
      imageName, newImageName
    integer(int16), intent(in), optional :: xStartArg, yStartArg

    type(DataSetT) :: inSet, outSet
    type(ArrayT) :: inArray, outArray
!    type(BlockT) :: inBlock, outBlock
!    type(DssT) ::&
!      my_dss
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    integer(int32), dimension(:), pointer ::&
      dimVector
    integer(int16) :: xStart, yStart, xFinis, yFinis

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    inSet = dataSet(imageName, READ)
    inArray = array(inSet, 0, READ)
!    inBlock = block(inArray)
    outSet = dataSet(newImageName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(newImage, 1)
    dimVector(1) = size(newImage, 2)
    xFinis = xStart + size(newImage, 1) - 1
    yFinis = yStart + size(newImage, 2) - 1
    outArray = addArray(outSet, 'PRIMARY', REAL32, dimVector, position=0)
!    outBlock = block(outArray)
    call setAttributes(attributable(outSet), attributable(inSet))

    call copyVitalArrayKwds(inArray, outArray)
    ! This is done instead of a total copy of all array-header keywords because
    ! doing the latter stuffs up the DSS.

    arrayDataReal32 => real32Array2Data(outArray)
    arrayDataReal32 = 0.0
    arrayDataReal32(xStart-1:xFinis-1, yStart-1:yFinis-1) = newImage

    call release(outSet)
    call release(inSet)

    deallocate(dimVector)

  end subroutine saveOutImageReal32_old

  !---------------------------------------------------------------------
  subroutine saveOutImageInt32(newImage, imageName, newImageName&
  , xStartArg, yStartArg)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) ::&
      newImage(:,:)
    character(*), intent(in) ::&
      imageName, newImageName
    integer(int16), intent(in), optional :: xStartArg, yStartArg

    type(DataSetT) :: inSet, outSet
    type(ArrayT) :: inArray, outArray
!    type(BlockT) :: inBlock, outBlock
!    type(DssT) ::&
!      my_dss
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    integer(int32), dimension(:), pointer ::&
      dimVector
    integer(int16) :: xStart, yStart, xFinis, yFinis

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    inSet = dataSet(imageName, READ)
    inArray = array(inSet, 0, READ)
!    inBlock = block(inArray)
    outSet = dataSet(newImageName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(newImage, 1)
    dimVector(1) = size(newImage, 2)
    xFinis = xStart + size(newImage, 1) - 1
    yFinis = yStart + size(newImage, 2) - 1
    outArray = addArray(outSet, 'PRIMARY', INTEGER32, dimVector, position=0)
!    outBlock = block(outArray)
    call setAttributes(attributable(outSet), attributable(inSet))

    call copyVitalArrayKwds(inArray, outArray)
    ! This is done instead of a total copy of all array-header keywords because
    ! doing the latter stuffs up the DSS.

    arrayDataInt32 => int32Array2Data(outArray)
    arrayDataInt32 = 0
    arrayDataInt32(xStart-1:xFinis-1, yStart-1:yFinis-1) = newImage

    call release(outSet)
    call release(inSet)

    deallocate(dimVector)

  end subroutine saveOutImageInt32

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module save_image
